import time
import requests
import threading
import queue
import signal
import fnmatch
import os
import csv
from rich.console import Console
from cryptography.x509.oid import NameOID


from phisher.config import *
from phisher.decoder import extract_cert
from phisher.telegram_notifier import notify_in_telegram



console = Console() 
print_lock = threading.Lock()


class CTMonitor:
    def __init__(self, keywords_file, output_file, format, keywords=None, notify=None, log=False, log_list_url='https://www.gstatic.com/ct/log_list/v3/log_list.json'):
        self.keywords_file = keywords_file
        self.keywords_list = keywords
        self.log = log
        self.output_file = output_file
        self.output_format = format
        self.log_list_url = log_list_url
        self.urls = self._get_urls()
        self.q = queue.Queue()
        self.last_tree_sizes = {}
        self.producer_thread = threading.Thread(target=self._producer)
        self.consumer_threads = [threading.Thread(target=self._consumer) for _ in range(2)]
        self.keywords = self._read_keywords()
        self.notify = notify

    def _write_csv(self, cn, cert):
        """ Writer to csv file """

        fields = ["Subject", "Issuer", "Serial No", "Version", "Not Before (UTC)", "Not After (UTC)"]
        data = {
            "Subject" : cert.subject.rfc4514_string(),
            "Issuer":cert.issuer.rfc4514_string(),
            "Serial No":cert.serial_number,
            "Version":cert.version.name, 
            "Not Before (UTC)":cert.not_valid_before_utc, 
            "Not After (UTC)":cert.not_valid_after_utc
        }
        with open(self.output_file, 'a', newline='') as f:
            writer = csv.DictWriter(f, fieldnames=fields)
            writer.writerow(data)

    def _write_txt(self, cn, cert):
        """ Writer to txt file """

        info = (
            f"[*] Found domain: {cn}\n"
            f"Subject:     {cert.subject.rfc4514_string()}\n"
            f"Issuer:      {cert.issuer.rfc4514_string()}\n"
            f"Serial No.:  {cert.serial_number}\n"
            f"Version:     {cert.version.name}\n"
            f"Validity:\n"
            f"  Not Before (UTC): {cert.not_valid_before_utc}\n"
            f"  Not After  (UTC): {cert.not_valid_after_utc}"
        )
        with open(self.output_file, 'a') as f:
            print(info, file=f)

    def _write_results(self, cn, cert):
        """ Logs certificate details to a file """
        
        if self.output_format == 'txt':
            self._write_txt(cn, cert)
        
        if self.output_format == 'csv':
            self._write_csv(cn, cert)
    
    def _read_keywords(self):
        """ Read keywords """
        k = []
        if os.path.exists(self.keywords_file):
            with open(self.keywords_file) as f:
                raw = [line.strip() for line in f if line.strip()]
            k.extend([f"*{kw.lower()}*" for kw in raw])
        if self.keywords_list:
            k.extend([ f"*{k}*" for k in self.keywords_list ])
        
        return k if k else ["*"]
                
        

    def _safe_print(self, *args, **kwargs):
        """ Function to print on time while multiple threads are used """
        
        with print_lock:
            console.print(*args, **kwargs, highlight=False)
    def _get_urls(self):
        """ Gets the list of current CT logs urls """

        resp = requests.get(self.log_list_url)
        data = resp.json()

        result = []
        for operator in data.get('operators', []):
            name = operator.get('name')
            logs = operator.get('logs', [])
            if not name or not logs:
                continue

            first_url = logs[0].get('url', '')
            if not first_url or 'example.com' in first_url:
                continue

            result.append(first_url)
        return result
    
    def _get_size(self, url):
        """ Get the current CT logs tree size """
        
        r = requests.get(f"{url}ct/v1/get-sth")
        return r.json()['tree_size']

    def _get_entries(self, url, start, end):
        """ Get the CT log entries for given url. start and end defien the startind and ending index """
        
        params = {'start': start, 'end': end}
        r = requests.get(f"{url}ct/v1/get-entries", params=params)
        if r.status_code == 200:
            return r.json()
        
    def _poll_once(self):
        """ Functions to read up to BATCH_SIZE many CT logs entries """

        for url in self.urls:
            size = self._get_size(url)
            prev = self.last_tree_sizes.get(url, size-1)

            if size > prev:
                for start in range(prev, size, BATCH_SIZE):
                    end = min(start + BATCH_SIZE - 1, size)
                    entries = self._get_entries(url, start, end).get('entries', [])
                    if not entries:
                        continue
                    else:
                        for e in entries:
                            self.q.put(e)
                    time.sleep(MIN_INTERVAL)
                self.last_tree_sizes[url] = size
            time.sleep(MIN_INTERVAL)
    
    def _producer(self):
        """ Producer function for multithreading """

        while True:
            self._poll_once()
    
    def _consumer(self):
        """ Consumer function for multithreading """

        while True:
            batch = []
            for _ in range(BATCH_SIZE_CONSUMER):
                try:
                    item = self.q.get(timeout=1)
                except queue.Empty:
                    break
                batch.append(item)

            if not batch:
                continue 

            for ct_entry in batch:
                try:
                    leaf_input = ct_entry["leaf_input"]
                    extra_data = ct_entry["extra_data"]
                    self._check_ct_entry(leaf_input, extra_data)
                except Exception as e:
                    self._safe_print("Error occured:", e)

            for _ in batch:
                self.q.task_done()
    def _contains_keywords(self, domain):
        """ Checks if domain contains the keywords """

        for pat in self.keywords:
            if fnmatch.fnmatchcase(domain.lower(), pat):
                return True
        return False

    def _check_ct_entry(self, leaf_input, extra_data):
        cert = extract_cert(leaf_input, extra_data)
        if not cert:
            return
        cn_attrs = cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME)
        common_names = [attr.value for attr in cn_attrs]

        for cn in common_names:
            if self._contains_keywords(cn):
                self._safe_print(f"[#2596be][*] Found domain:[/#2596be] {cn}")
                if self.notify:
                    notify_in_telegram(f"Found domain {cn}", self.notify[0], self.notify[1])
                if self.log:
                    self._write_results(cn, cert)             
                
    def start(self):
        t_prod = threading.Thread(target=self._producer, args=())
        t = threading.Thread(target=self._consumer, args=())
            
        t_prod.start()
        t.start()
            
        signal.signal(signal.SIGINT, signal.SIG_DFL) 

        t_prod.join()  
        t.join()
        self.q.join()  


    


import argparse
from phisher.monitor import *
from rich.console import Console
from rich.text import Text
from rich_gradient import Gradient
from rich_argparse import RichHelpFormatter

console = Console()


def parse_args():
    p = argparse.ArgumentParser(
        prog="Phisher",
        description="A tool for monitoring the certificate transparacy logs.",
        formatter_class=RichHelpFormatter
    )
    p.add_argument(
        "--notify", "-n",
        nargs=2,                         
        metavar=("CHAT_ID", "BOT_TOKEN"), 
        help="Enable notifications: provide CHAT_ID and BOT_TOKEN"
    )
    p.add_argument(
        "--log", "-l",
        action="store_true",
        help="Enable logging results to a file. By default will log to a domains.txt file."
    )
    p.add_argument(
        "--keywords", "-k",
        nargs="+",
        metavar="KEYWORDS",
        help="Space separated list of keywords"
    )
    p.add_argument(
        "--keywords-file", "-kf",
        default="keywords.txt",
        help="Path to newline-separated keywords file"
    )

    p.add_argument(
        "--format", "-f",
        choices=["csv", "txt"],
        default="txt",
        help="Choose the output format: 'csv' or 'txt'."
    )
    p.add_argument(
        "--output-file", "-o",
        default="domains.txt",
        help="File to append matching domain reports to"
    )
    return p.parse_args()

def main():
    args = parse_args()
    raw_banner = """
██████╗ ██╗  ██╗██╗███████╗██╗  ██╗███████╗██████╗
██╔══██╗██║  ██║██║██╔════╝██║  ██║██╔════╝██╔══██╗
██████╔╝███████║██║███████╗███████║█████╗  ██████╔╝
██╔═══╝ ██╔══██║██║╚════██║██╔══██║██╔══╝  ██╔══██╗
██║     ██║  ██║██║███████║██║  ██║███████╗██║  ██║
╚═╝     ╚═╝  ╚═╝╚═╝╚══════╝╚═╝  ╚═╝╚══════╝╚═╝  ╚═╝
    """

    colors = ["#01ff87", "#01ffd6","#01f7ff", "#008eff", "#628bff","#5e5299", "purple4","dark_magenta","dark_magenta" ,"dark_magenta" , "purple3" , "purple3" , "purple3" ,"purple3" , "deep_pink3" ,"deep_pink3" ]
    console.print(Gradient(raw_banner, colors=colors))
    log = args.log or args.output_file
    monitor = CTMonitor(args.keywords_file, args.output_file, args.format,args.keywords ,args.notify, log )
    monitor.start()

if __name__ == '__main__':
    main()


from telegram import Bot
import asyncio

def notify_in_telegram(message, chat_id, bot_token):
    """ Function to notify in telegram """
        
    bot = Bot(token=bot_token)

    async def send_message(text, chat_id):
        async with bot:
            await bot.send_message(text=text, chat_id=chat_id)

    async def run_bot(messages, chat_id=chat_id):
        await send_message(messages, chat_id)


    def send_sync(message):
        asyncio.run(run_bot(message, chat_id))

    send_sync(message)


import base64
from construct import Struct, Byte, Int16ub, Int64ub, Enum, Bytes, Int24ub, this, GreedyBytes, GreedyRange
from cryptography import x509
from cryptography.hazmat.backends import default_backend

MerkleTreeHeader = Struct(
	"Version"		 / Byte,
	"MerkleLeafType"  / Byte,
	"Timestamp"	   / Int64ub,
	"LogEntryType"	/ Enum(Int16ub, X509LogEntryType=0, PrecertLogEntryType=1),
	"Entry"		   / GreedyBytes
)

Certificate = Struct(
	"Length" / Int24ub,
	"CertData" / Bytes(this.Length)
)

CertificateChain = Struct(
	"ChainLength" / Int24ub,
	"Chain" / GreedyRange(Certificate),
)

PrecertChainEntry = Struct(
    "Precert"        / Certificate,
    "IssuerKeyHash"  / Bytes(32),  
    "ChainLength"    / Int24ub,           
    "Chain"          / GreedyRange(Certificate),
)

def extract_cert(leaf_input, extra_data):
    """ Extracts the certificated from leaf_input and extra_data from CT logs entry """
    
    raw_leaf = base64.urlsafe_b64decode(leaf_input + "===")
    hdr = MerkleTreeHeader.parse(raw_leaf)

    if hdr.LogEntryType == "X509LogEntryType":
        der = Certificate.parse(hdr.Entry).CertData
    else:
        raw_extra = base64.urlsafe_b64decode(extra_data + "===")
        precert_blob = PrecertChainEntry.parse(raw_extra)
        der = precert_blob.Precert.CertData
        
    cert = x509.load_der_x509_certificate(der, default_backend())
    return cert



import setuptools
import os

long_description = ""
if os.path.isfile("README.md"):
    with open("README.md", "r", encoding="utf-8") as fh:
        long_description = fh.read()

setuptools.setup(
    name="phisher-ctmonitor",
    version="0.1.0",
    author="Viktoryia Valiuk",
    author_email="viktoriapogorzhelska@gmail.com",
    description="A tool for monitoring the certificate transparency logs.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nbdknws/phisher",
    packages=setuptools.find_packages(),
    install_requires=[
        "construct==2.10.70",
        "cryptography==45.0.4",
        "python-telegram-bot==22.2",
        "Requests==2.32.4",
        "rich==14.0.0",
        "rich_argparse==1.7.1",
        "rich_gradient==0.3.2",
    ],
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
        # phisher will invoke the main() in your __main__.py (or wherever you point it)
        "phisher = phisher.__main__:main",
        ],
    },
)